# Copyright 2023 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# curate.data_transform.py
"""Data Transformations"""
import logging
import re

import numpy as np
import pandas as pd
import patsy as pt

from ..utils.snakey_lowercaser import SnakeyLowercaser

__all__ = [
    'DatatypeConverter',
    'DatasetReshaper',
    'Transformer',
    'Standardizer',
    'compress_factor_levels',
]

_log = logging.getLogger(__name__)


class DatatypeConverter:
    """Force correct datatypes according to what model expects"""

    def __init__(self, ftsd: dict, ftslvlcat: dict = {}, date_format: str = '%Y-%m-%d'):
        """Initialise with fts and optionally specify factors with specific levels

        Use with a fts dict of form:
            ftsd = dict(
                fcat = [],
                fstr = [],
                fbool = [],
                fdate = [],
                fyear = [],
                fint = [],
                ffloat = [],
                fverbatim = [],        # maintain in current dtype)
        """
        self.ftsd = dict(
            fcat=ftsd.get('fcat', []),
            fstr=ftsd.get('fstr', []),
            fbool=ftsd.get('fbool', []),
            fdate=ftsd.get('fdate', []),
            fyear=ftsd.get('fyear', []),
            fint=ftsd.get('fint', []),
            ffloat=ftsd.get('ffloat', []),
            fverbatim=ftsd.get('fverbatim', []),  # keep verbatim
        )
        self.ftslvlcat = ftslvlcat
        self.rx_number_junk = re.compile(r'[#$€£₤¥,;%\s]')
        self.date_format = date_format
        inv_bool_dict = {
            True: ['yes', 'y', 'true', 't', '1', 1, 1.0],
            False: ['no', 'n', 'false', 'f', '0', 0, 0.0],
        }
        self.bool_dict = {v: k for k, vs in inv_bool_dict.items() for v in vs}
        self.strnans = ['none', 'nan', 'null', 'na', 'n/a', 'missing', 'empty', '']

    def _force_dtypes(self, dfraw):
        """Select fts and convert dtypes. Return cleaned df"""
        snl = SnakeyLowercaser()

        # subselect desired fts
        # TODO make this optional
        fts_all = [w for _, v in self.ftsd.items() for w in v]
        df = dfraw[fts_all].copy()

        for ft in self.ftsd['fcat'] + self.ftsd['fstr']:
            # tame string, clean, handle nulls
            idx = df[ft].notnull()
            vals = df.loc[idx, ft].astype(str, errors='raise').apply(snl.clean)
            df.drop(ft, axis=1, inplace=True)
            df.loc[~idx, ft] = 'nan'
            df.loc[idx, ft] = vals
            if ft in self.ftsd['fcat']:
                df[ft] = pd.Categorical(df[ft].values, ordered=True)
            else:
                df[ft] = df[ft].astype('string')

        for ft in self.ftsd['fbool']:
            # tame string, strip, lower, use self.bool_dict, use pd.NA
            if df.dtypes[ft] == object:
                df[ft] = df[ft].apply(lambda x: str(x).strip().lower())
                df.loc[df[ft].isin(self.strnans), ft] = np.nan
                df[ft] = df[ft].apply(lambda x: self.bool_dict.get(x, x))

                if set(df[ft].unique()) != set([True, False, np.nan]):
                    # avoid converting anything not yet properly mapped
                    continue
            df[ft] = df[ft].convert_dtypes(convert_boolean=True)

        for ft in self.ftsd['fyear']:
            if df.dtypes[ft] == object:
                df[ft] = (
                    df[ft]
                    .astype(str)
                    .str.strip()
                    .str.lower()
                    .map(lambda x: self.rx_number_junk.sub('', x))
                )
                df.loc[df[ft].isin(self.strnans), ft] = np.nan
            df[ft] = df[ft].astype(float, errors='raise')
            if pd.isnull(df[ft]).sum() == 0:
                df[ft] = df[ft].astype(int, errors='raise')
            df[ft] = pd.to_datetime(df[ft], errors='raise', format='%Y')

        for ft in self.ftsd['fdate']:
            df[ft] = pd.to_datetime(df[ft], errors='raise', format=self.date_format)
        try:
            for ft in self.ftsd['fint']:
                if df.dtypes[ft] == object:
                    df[ft] = (
                        df[ft]
                        .astype(str)
                        .str.strip()
                        .str.lower()
                        .map(lambda x: self.rx_number_junk.sub('', x))
                    )
                    df.loc[df[ft].isin(self.strnans), ft] = np.nan
                df[ft] = df[ft].astype(float, errors='raise')
                if pd.isnull(df[ft]).sum() == 0:
                    df[ft] = df[ft].astype(int, errors='raise')
        except Exception as e:
            raise Exception(f'{str(e)} in ft: {ft}').with_traceback(
                e.__traceback__
            )  # from e

        try:
            for ft in self.ftsd['ffloat']:
                if df.dtypes[ft] == object:
                    df[ft] = (
                        df[ft]
                        .astype(str)
                        .str.strip()
                        .str.lower()
                        .map(lambda x: self.rx_number_junk.sub('', x))
                    )
                    df.loc[df[ft].isin(self.strnans), ft] = np.nan
                df[ft] = df[ft].astype(float, errors='raise')
        except Exception as e:
            raise e(ft)

        # NOTE verbatim will simply remain. We're now at the end of the columns
        # for ft in self.fts['fverbatim']:
        #     _log.info(f'Kept ft verbatim: {ft}')

        return df

    def convert_dtypes(self, df):
        """Force dtypes for recognised features (fts) in df"""
        dfclean = self._force_dtypes(df)

        for ft, lvls in self.ftslvlcat.items():
            dfclean[ft] = pd.Categorical(
                dfclean[ft].values, categories=lvls, ordered=True
            )

        return dfclean


class DatasetReshaper:
    """Convenience functions to reshape whole datasets"""

    def __init__(self):
        pass

    def create_dfcmb(self, df: pd.DataFrame, fts: dict) -> pd.DataFrame:
        """Create a combination dataset `dfcmb` from inputted `df`.

        *Not* a big groupby (equiv to cartesian join) for factor values
        and concats numerics. Instead, this concats unique vals in cols
        which yields a far more compact dfcmb. Note that the columns will
        be ragged so need to fill NULLS with (any) value from that column

        The shape and datatypes matter.

        We use this because as of the current patsy (0.5.1), design_info
        objects still can't be pickled, which means that when you want to
        transform a test / holdout sample according to a training dataset
        you have to instantiate a new design_info object in memory.

        In this codebase we set the design_info attribute on the
        Transformer object, so you can keep that object around in memory,
        but upon memory-loss you need to call it with the dfcmb created
        here e.g. Transformer.fit_transform(dfcmb).

        I recommend storing dfcmb in a DB: it only needs to be updated if
        the modelled features change (e.g. include a new feature) or if
        factor values change (e.g. a new construction_type).

        NOTE: only takes datatypes as understood by patsy: factor-values
            (aka categoricals aka strings), or ints or floats. No dates.
        """
        dfcmb = pd.DataFrame(index=[0])
        fts_factor = fts.get('fcat', []) + fts.get('fbool', [])
        for ft in fts_factor:
            colnames_pre = list(dfcmb.columns.values)
            s = pd.Series(np.unique(df[ft]), name=ft)
            dfcmb = pd.concat([dfcmb, s], axis=1, join='outer', ignore_index=True)
            dfcmb.columns = colnames_pre + [ft]

        for ft in dfcmb.columns:
            dfcmb[ft] = dfcmb[ft].fillna(dfcmb[ft][:1].values[0])
            dfcmb[ft] = dfcmb[ft].astype(df[ft].dtype)

            # TODO: force order for categorical
            # df['fpc_aais_ctgry'] = pd.Categorical(df['fpc_aais_ctgry'].values, categories=vals, ordered=True)

        for ft in fts.get('fint'):
            dfcmb[ft] = 1

        for ft in fts.get('ffloat'):
            dfcmb[ft] = 1.0

        _log.info(
            'Reduced df ({} rows, {:,.0f} KB) to dfcmb ({} rows, {:,.0f} KB)'.format(
                df.shape[0],
                df.values.nbytes / 1e3,
                dfcmb.shape[0],
                dfcmb.values.nbytes / 1e3,
            )
        )

        return dfcmb

    def _create_dfcmb_big(self, df: pd.DataFrame, fts: dict) -> pd.DataFrame:
        """Create a combination dataset `dfcmb` from inputted `df`.
        Just a big groupby (equiv to cartesian join) for factor values
        and concats numerics. The shape and datatypes matter.

        We use this because as of the current patsy (0.5.1), design_info
        objects still can't be pickled, which means that when you want to
        transform a test / holdout sample according to a training dataset
        you have to instantiate a new design_info object in memory.

        In this codebase we set the design_info attribute on the
        Transformer object, so you can keep that object around in memory,
        but upon memory-loss you need to call it with the dfcmb created
        here e.g. Transformer.fit_transform(dfcmb).

        I recommend storing dfcmb in a DB: it only needs to be updated if
        the modelled features change (e.g. include a new feature) or if
        factor values change (e.g. a new construction_type).

        NOTE: only takes datatypes as understood by patsy: factor-values
            (aka categoricals aka strings), or ints or floats. No dates.
        """
        dfcmb = pd.DataFrame(index=[0])
        fts_factor = fts.get('fcat', []) + fts.get('fbool', [])
        if len(fts_factor) > 0:
            dfcmb = df.groupby(fts_factor).size().reset_index().iloc[:, :-1]

        for ft in fts.get('fint'):
            dfcmb[ft] = 1

        for ft in fts.get('ffloat'):
            dfcmb[ft] = 1.0

        _log.info(
            'Reduced df ({} rows, {:,.0f} KB) to dfcmb ({} rows, {:,.0f} KB)'.format(
                df.shape[0],
                df.values.nbytes / 1e3,
                dfcmb.shape[0],
                dfcmb.values.nbytes / 1e3,
            )
        )

        return dfcmb


class Transformer:
    """Model-agnostic patsy transformer from row-wise natural observations
    into dmatrix according to patsy formula or patsy design_info object
    NOTES:
        + design_info is stateful
        + it's reasonable to initialise this per-observation but far more
          efficient to initialise once and persist in-memory
        + allows for F() and F():F() terms
    """

    def __init__(self):
        self.design_info = None
        self.col_idx_numerics = None
        self.rx_get_f = re.compile(r'(F\(([a-z0-9_:]+?)\))')
        self.fts_fact_mapping = {}
        self.original_fml = None

    def fit_transform(
        self, fml: str, df: pd.DataFrame, propagate_nans: bool = False
    ) -> pd.DataFrame:
        """Fit a new design_info attribute for this instance according to
        `fml` acting upon `df`. Return the transformed dmatrix (np.array)
        Use this for a new training set or to initialise the transformer
        based on dfcmb.
        """
        self.original_fml = fml
        # deal w/ any fml components F(), use set() to uniqify
        fts_f = list(set(self.rx_get_f.findall(fml)))

        if len(fts_f) > 0:
            df = df.copy()
            for ft_f in fts_f:
                dt = df[ft_f[1]].dtype.name
                if dt != 'category':
                    raise AttributeError(
                        f'fml contains F({ft_f[1]}),'
                        + f' dtype={dt}, but {ft_f[1]} is not categorical'
                    )
                # map feature to int based on its preexisting categorical order
                # https://stackoverflow.com/a/55304375/1165112
                map_int_to_fact = dict(enumerate(df[ft_f[1]].cat.categories))
                map_fact_to_int = {v: k for k, v in map_int_to_fact.items()}
                self.fts_fact_mapping[ft_f[1]] = map_fact_to_int
                df[ft_f[1]] = df[ft_f[1]].map(map_fact_to_int).astype(np.int)

                # replace F() in fml so patsy can work as normal w/ our new int type
                fml = fml.replace(ft_f[0], ft_f[1])
        _log.info(f'Created fml: {fml}')
        # TODO add option to output matrix   # np.asarray(mx_ex)
        # TODO add check for fml contains `~` and handle accordingly

        # do nothing, see https://stackoverflow.com/a/51641183/1165112
        na_action = pt.NAAction(NA_types=[]) if propagate_nans else 'raise'

        df_ex = pt.dmatrix(fml, df, NA_action=na_action, return_type='dataframe')
        self.design_info = df_ex.design_info

        # force patsy transform of an F() to int feature back to int not float
        fts_force_to_int = ['intercept']  # also force intercept
        fts_force_to_int = list(self.fts_fact_mapping.keys())
        if len(fts_force_to_int) > 0:
            df_ex[fts_force_to_int] = df_ex[fts_force_to_int].astype(np.int64)

        return df_ex

    def transform(self, df: pd.DataFrame, propagate_nans: bool = False) -> pd.DataFrame:
        """Transform input `df` to dmatrix according to pre-fitted
        `design_info`. Return transformed dmatrix (np.array)

        NEW FUNCTIONALITY: 2021-03-11
            factorize components marked as F(), must be pd.Categorical
        """
        if self.design_info is None:
            raise AttributeError('No design_info, run `fit_transform()` first')

        # # hacky way to get F():F() components
        # fts_ff = set(self.rx_get_ff.findall(self.original_fml))

        # map any features noted in fts_fact_mapping
        try:
            df = df.copy()
            for ft, map_fact_to_int in self.fts_fact_mapping.items():
                df[ft] = df[ft].map(map_fact_to_int).astype(np.int64)
        except AttributeError:
            # self.fts_fact_mapping was never created for this instance
            # simply because no F() in fml
            pass

        # TODO add option to output matrix
        # mx_ex = pt.dmatrix(self.design_info, df, NA_action='raise',
        #                       return_type='matrix')
        # return np.asarray(mx_ex)

        # do nothing, see https://stackoverflow.com/a/51641183/1165112
        na_action = pt.NAAction(NA_types=[]) if propagate_nans else 'raise'

        df_ex = pt.dmatrix(
            self.design_info, df, NA_action=na_action, return_type='dataframe'
        )
        self.design_info = df_ex.design_info

        # force patsy transform of an index feature back to int!
        # there might be a better way to do this
        fts_force_to_int = []
        fts_force_to_int = list(self.fts_fact_mapping.keys())
        if len(fts_force_to_int) > 0:
            df_ex[fts_force_to_int] = df_ex[fts_force_to_int].astype(np.int64)

        return df_ex


class Standardizer:
    """Model-agnostic standardizer from pre-transformed dmatrix
    dmatrix according to patsy formula or patsy design_info object
    NOTES:
    + means, sdevs and scale are stateful: anticipate never changing
    + must be initialised with Transformer.design_info, consider refactoring
    + it's reasonable to initialise this per-observation but far more
        efficient to initialise once and persist in-memory
    + note these standardizations use nanmean, nanstd, so they're
        already compatible with recent changes (2021-03-31) to
        transformer to propagate nans

    NEW FUNCTIONALITY:
    + apply standardization using a mask. allows us to exclude any col
    + rework to I/O dataframes

    TODO: introduce minmax scaling as an option
    """

    def __init__(self, design_info: pt.design_info.DesignInfo, fts_exclude: list = []):
        """Optionally exclude from standardization a list of named fts that
        are numeric and would otherwise get standardardized"""

        self.design_info = design_info
        col_num_excl = [0] + [
            i
            for i, n in enumerate(self.design_info.column_names)
            if (n in fts_exclude) or re.search(r'\[T\.', n)
        ]

        # col_mask is True where we want to exclude the col from standardization
        self.col_mask = [
            True if i in col_num_excl else False
            for i in np.arange(len(self.design_info.column_names))
        ]

        self.means = None  # will become np.ndarray
        self.sdevs = None  # will become np.ndarray
        self.scale = None  # will become int

    def standardize(self, df: pd.DataFrame) -> pd.DataFrame:
        """Standardize input df to mean-centered, 2sd unit variance,
        Retain the fitted means and sdevs for later use in standardize()
        """
        if any([v is None for v in [self.means, self.sdevs, self.scale]]):
            raise AttributeError(
                'No mns, sdevs or scale, '
                + 'run `fit_standardize()` on training set first'
            )

        df_s = (df - self.means) / (self.sdevs * self.scale)
        mask = np.tile(self.col_mask, (len(df), 1))

        # fill original df w/ standardized to more easily preseve dtype of ints
        df_exs = df.mask(~mask, df_s)  # replace values where condition is True
        return df_exs

    def fit_standardize(self, df: pd.DataFrame, scale: int = 2) -> pd.DataFrame:
        """Standardize numeric features of df with variable scale
        Retain the fitted means and sdevs for later use in standardize()
        """
        self.means = np.where(self.col_mask, np.nan, np.nanmean(df, axis=0))
        self.sdevs = np.where(self.col_mask, np.nan, np.nanstd(df, axis=0))
        self.scale = scale
        return self.standardize(df)

    def standardize_mx(self, mx: np.ndarray) -> np.ndarray:
        """Standardize input mx to mean-centered, 2sd unit variance,
        Retain the fitted means and sdevs for later use in standardize()
        """
        if any([v is None for v in [self.means, self.sdevs, self.scale]]):
            raise AttributeError(
                'No mns, sdevs or scale, '
                + 'run `fit_standardize()` on training set first'
            )

        mxs_all = (mx - self.means) / (self.sdevs * self.scale)
        mask = np.tile(self.col_mask, (len(mx), 1))
        mxs = np.where(mask, mx, mxs_all)
        return mxs

    def fit_standardize_mx(self, mx: np.ndarray, scale: int = 2) -> np.ndarray:
        """Standardize numeric features of mx with variable scale
        Retain the fitted means and sdevs for later use in standardize()
        """
        # TODO add option to output dataframe
        self.means = np.where(self.col_mask, np.nan, np.nanmean(mx, axis=0))
        self.sdevs = np.where(self.col_mask, np.nan, np.nanstd(mx, axis=0))
        self.scale = scale
        return self.standardize(mx)

    def get_scale(self):
        """Get values followuing fit_standardize. Persist values over time."""
        means_sdevs = pd.DataFrame(
            {'means': self.means, 'sdevs': self.sdevs},
            index=self.design_info.column_names,
        )

        return means_sdevs, self.scale

    def set_scale(self, means_sdevs, scale: int = 2):
        """Set values saved from a prior fit_standardize. Now can run
        standardize for new data
        """
        self.means = means_sdevs['means'].values
        self.sdevs = means_sdevs['sdevs'].values
        self.scale = scale


def compress_factor_levels(df: pd.DataFrame, fts: list, topn: int = 20) -> pd.DataFrame:
    """Crude compression for factor levels, into the topn + 1 (other)
    Return new dataframe for fts
    """
    newdf = pd.DataFrame(index=df.index)
    for ft in fts:
        vc = df[ft].value_counts(dropna=False)
        _log.info(f'{ft}: compress {vc[:topn].sum()} ({vc[:topn].sum()/vc.sum():.1%})')
        _log.info(
            f'{ft}: compressed {len(vc)}-{topn} ({len(vc)-topn}) levels '
            + 'into `other`, '
            + f'{vc[topn:].sum()} rows ({vc[topn:].sum() / len(df):.1%}) affected'
        )
        vc_map = {
            k: (k if i < topn else 'other')
            for i, (k, v) in enumerate(vc.to_dict().items())
        }
        newdf[f'{ft}_topn'] = df[ft].map(vc_map)

    return newdf



# Copyright 2023 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model.describe.py
"""Model Descriptions"""
import re

import arviz as az
import numpy as np
import pandas as pd
import patsy as pat

from ..model_pymc import BasePYMCModel

# import pytensor.tensor as pt
# from IPython.display import Markdown, display


__all__ = [
    'model_desc',
    'extract_yobs_yhat',
    'describe_dist',
    'get_summary',
]  # , 'print_rvs']

RSD = 42
rng = np.random.default_rng(seed=RSD)


def model_desc(fml: str) -> str:
    """Convenience: return patsy modeldesc
    NOTE: `.describe()` doesn't return the `1 +` (intercept) term in the
        case that it's present. check and add if needed
    """
    fmls = fml.split(' ~ ')
    add_intercept = False if re.match(r'1 \+', fml) is None else True
    r = pat.ModelDesc.from_formula(fml).describe()
    if len(fmls) == 2:
        rs = r.split(' ~ ')
        if add_intercept:
            r = f'{rs[0]} ~ 1 + {rs[1]}'
    elif len(fmls) == 1:
        if add_intercept:
            r = f'1 + {r[2:]}'
    else:
        raise ValueError('fml must have only a single tilde `~`')
    return f'patsy linear model desc:\n{r}\n'


def extract_yobs_yhat(
    idata: az.InferenceData, obs: str = 'y', pred: str = 'yhat'
) -> tuple:
    """Convenience: extract y_obs, y_hat from idata
    get yhat in the shape (nsamples, nobs)
    """
    nsamp = np.product(idata.posterior_predictive[pred].shape[:-1])
    yobs = idata.constant_data[obs].values  # (nobs,)
    yhat = idata.posterior_predictive[pred].values.reshape(nsamp, -1)  # (nsamp, nobs)
    return yobs, yhat


def describe_dist(mdl: BasePYMCModel, log: bool = False, inc_summary: bool = False):
    """Convenience: get distribution descriptions from distributions.DistNumpy
    and return for printing or Markdown
    NOTE: consider deprecating
    """
    title = f'{mdl.name}: Natural Distributions'
    dist = mdl.dist_natural
    if log:
        title = f'{mdl.name}: Logged Distributions'
        dist = mdl.dist_log

    if inc_summary:
        return title, {**mdl.notation, **dist, **mdl.conditions, **mdl.summary_stats}
    return title, {**dist}


def get_summary(mdl: BasePYMCModel, rvs: list, group='posterior') -> pd.DataFrame:
    """Convenience fn to get arviz summary of idata posteriors"""

    df = az.summary(mdl.idata, var_names=rvs, group=group)
    return df


def print_rvs(mdl: BasePYMCModel) -> list[str]:
    """Convenience to print RV strings to notebook
    Use as _ = [display(Markdown(s for s in mt.print_rvs(mdl)))]
    """
    return [
        rv.str_repr(formatting='string', include_params=True)
        for rv in mdl.model.free_RVs + mdl.model.potentials
    ]


# _ = [print(f'{k}: {v}') for k, v in mdl.describe_rvs().items()]

# def print_rvs(rvs: list[pt.TensorVariable]) -> None:
#     """Display rvs to Notebook using latex, post sub underscores
#     Hack to replace print(mdl.model) see https://github.com/pymc-devs/pymc/issues/6869
#     """
#     rx_name_usc = re.compile(r"(?:text\{[^\_\}]+?)(\_+)(?:[^\_\}]+?)(?:(\_+)(?:[^\_\}]+?))*?(?:\})", re.I)
#     for rv in rvs:
#         s = rv.str_repr(formatting='latex', include_params=True)
#         t = rx_name_usc.sub(r"\\_", s)
#         display(Markdown(t))



# Copyright 2023 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# eda.eda_io.py
"""EDA File Handling"""
import logging
from pathlib import Path

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib import figure

from ..utils.file_io import BaseFileIO
from .describe import describe, get_fts_by_dtype

__all__ = ['FigureIO', 'display_image_file', 'output_data_dict']

_log = logging.getLogger(__name__)


class FigureIO(BaseFileIO):
    """Helper class to save matplotlib.figure.Figure objects to image file"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def write(self, f: figure.Figure, fn: str, *args, **kwargs) -> Path:
        """Accept figure.Figure & fqn e.g. `plots/plot.png`, write to fqn"""
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix('.png'))
        f.savefig(
            fname=fqn, format='png', bbox_inches='tight', dpi=300, *args, **kwargs
        )
        _log.info(f'Written to {str(fqn.resolve())}')
        return fqn


def display_image_file(
    fqn: str, title: str = None, figsize: tuple = (12, 6)
) -> figure.Figure:
    """Hacky way to display pre-created image file in a Notebook
    such that nbconvert can see it and render to PDF
    Force to max width 16 inches, for fullwidth render in live Notebook and PDF

    NOTE:
    Alternatives are bad
        1. This one is entirely missed by nbconvert at render to PDF
        # <img src="img.jpg" style="float:center; width:900px" />

        2. This one causes following markdown to render monospace in PDF
        # from IPython.display import Image
        # Image("./assets/img/oreum_eloss_blueprint3.jpg", retina=True)
    """
    img = mpimg.imread(fqn)
    f, axs = plt.subplots(1, 1, figsize=figsize)
    _ = axs.imshow(img)
    ax = plt.gca()
    _ = ax.grid(False)
    _ = ax.set_frame_on(False)
    _ = plt.tick_params(
        top=False,
        bottom=False,
        left=False,
        right=False,
        labelleft=False,
        labelbottom=False,
    )
    if title is not None:
        _ = f.suptitle(f'{title}', y=1.0)
    _ = f.tight_layout()
    return f


def output_data_dict(
    df: pd.DataFrame, dd_notes: dict[str, str], fqp: Path, fn: str = ''
):
    """Convenience fn: output data dict"""

    # flag if is index
    idx_names = list(df.index.names)
    dfi = pd.DataFrame({'ft': idx_names, 'is_index': [True] * len(idx_names)})

    # get desc overview
    nrows = 3
    dfd = describe(df, nrows=nrows, return_df=True)
    cols = dfd.columns.values
    cols[:nrows] = [f'example_row_{i}' for i in range(nrows)]
    dfd.columns = cols

    # attach
    dfd = pd.merge(dfi, dfd, how='right', on='ft')
    dfd['is_index'] = dfd['is_index'].fillna(False)
    dfd.set_index('ft', inplace=True)

    # set dtypes categorical
    df_dtypes = get_fts_by_dtype(df.reset_index(), as_dataframe=True)
    dfd['dtype'] = df_dtypes['dtype']
    del df_dtypes

    # attached notes
    df_dd_notes = pd.DataFrame(dd_notes, index=['notes']).T
    df_dd_notes.index.name = 'ft'
    dfd = pd.merge(dfd, df_dd_notes, how='left', left_index=True, right_index=True)

    # write overview
    fileio = BaseFileIO(rootdir=fqp)
    fn = f'_{fn}' if fn != '' else fn
    fqn = fileio.get_path_write(f'datadict{fn}.xlsx')

    writer = pd.ExcelWriter(str(fqn), engine='xlsxwriter')
    dfd.to_excel(writer, sheet_name='overview', index=True)

    # write cats to separate sheets for levels (but not indexes since they're unique)
    for ft in dfd.loc[dfd['dtype'].isin(['categorical', 'cat'])].index.values:
        if ft not in df.index.names:
            dfg = (df[ft].value_counts(dropna=False) / len(df)).to_frame('prop')
            dfg.index.name = 'value'
            dfg.reset_index().to_excel(
                writer,
                sheet_name=f'{ft[:28]}...' if len(ft) >= 31 else ft,
                index=False,
                float_format='%.3f',
                na_rep='NULL',
            )

    writer.close()
    _log.info(f'Written to {str(fqn.resolve())}')
    return fqn



# Copyright 2023 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model_pymc.pymc_io.py
"""Handling of Model Posterior Samples"""
import logging
from pathlib import Path

import arviz as az
from pymc.model_graph import model_to_graphviz

from ..utils.file_io import BaseFileIO
from . import BasePYMCModel

__all__ = ['PYMCIO']

_log = logging.getLogger(__name__)


class PYMCIO(BaseFileIO):
    """Helper class to read/write NetCDF files for Arviz inference data.
    Can also write model graphs to file
    Note similar behaviour to curate.data_io.SimpleStringIO
    """

    def __init__(self, *args, **kwargs):
        """Inherit super"""
        super().__init__(*args, **kwargs)

    def read_idata(self, fn: str) -> az.InferenceData:
        """Read arviz.InferenceData object from fn e.g. `mdl.netcdf`"""
        fqn = self.get_path_read(Path(self.snl.clean(fn)).with_suffix('.netcdf'))
        idata = az.from_netcdf(str(fqn.resolve()))
        _log.info(f'Read model idata from {str(fqn.resolve())}')
        return idata

    def write_idata(self, mdl: BasePYMCModel, fn: str = '') -> Path:
        """Accept BasePYMCModel object and fn e.g. `mdl.netcdf`, write to file"""
        fn = f'idata_{mdl.name}.netcdf' if fn == '' else fn
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix('.netcdf'))
        mdl.idata.to_netcdf(str(fqn.resolve()))
        _log.info(f'Written to {str(fqn.resolve())}')
        return fqn

    def write_graph(
        self, mdl: BasePYMCModel, fn: str = '', fmt: str = 'png', **kwargs
    ) -> Path:
        """Accept a BasePYMCModel object mdl, get the graphviz representation
        Write to file and return the fqn to allow use within eda.display_image_file()
        """
        t = kwargs.pop('txtadd', None)
        fn = f"{'_'.join(filter(None, [mdl.name, t]))}.{fmt}" if fn == '' else fn
        fqn = self.get_path_write(fn)
        gv = model_to_graphviz(mdl.model, formatting='plain')
        if fmt == 'png':
            gv.attr(dpi='300')
        elif fmt == 'svg':
            pass
        else:
            raise ValueError('format must be in {"png", "svg"}')

        # gv auto adds the file extension, so pre-remove if present
        gv.render(filename=str(fqn.with_suffix('')), format=fmt, cleanup=True)
        _log.info(f'Written to {str(fqn.resolve())}')
        return fqn



# Copyright 2023 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# eda.describe.py
"""Data Descriptions"""
import logging

import numpy as np
import pandas as pd
from IPython.display import display
from scipy import stats

__all__ = ['describe', 'display_fw', 'display_ht', 'get_fts_by_dtype']

_log = logging.getLogger(__name__)

RSD = 42
RNG = np.random.default_rng(seed=RSD)


def describe(
    df: pd.DataFrame,
    nobs: int = 3,
    nfeats: int = 30,
    limit: int = 50,  # MB
    get_mode: bool = False,
    get_counts: bool = True,
    reset_index: bool = True,
    return_df: bool = False,
    **kwargs,
) -> str:
    """Concat transposed topN rows, numerical desc & dtypes
    Beware a dataframe full of bools or categoricals will error
    thanks to pandas.describe() being too clever
    Assume df has index.
    """

    len_idx = df.index.nlevels
    note = ''
    if nfeats + len_idx < df.shape[1]:
        note = 'NOTE: nfeats + index shown {} < width {}'.format(
            nfeats + len_idx, df.shape[1]
        )
    nbytes = df.values.nbytes
    _log.info(f'Shape: {df.shape}')
    _log.info(f'Memsize: {nbytes // 1e6:,.1f} MB')
    _log.info(f'Index levels: {df.index.names}')
    _log.info(f'{note}')

    limit *= 1e6
    if df.values.nbytes > limit:
        return f'Array memsize {nbytes // 1e6:,.1f} MB > {limit / 1e6:,.1f} MB limit'

    df = df.copy()
    if reset_index:
        nfeats += len(df.index.names)
        df = df.reset_index()

    # start with pandas describe, add on dtypes
    dfdesc = df.describe(include='all').T

    dfout = pd.concat((dfdesc, df.dtypes), axis=1, join='outer', sort=False)
    dfout = dfout.loc[df.columns.values]
    dfout.rename(columns={0: 'dtype', 'unique': 'count_unique'}, inplace=True)
    dfout.index.name = 'ft'

    # add counts for all
    if get_counts:
        dfout['count_notnull'] = df.shape[0] - df.isnull().sum()
        dfout['count_null'] = df.isnull().sum(axis=0)
        dfout['count_inf'] = (
            np.isinf(df.select_dtypes(np.number)).sum().reindex(df.columns)
        )
        dfout['count_zero'] = (
            (df.select_dtypes(np.number) == 0).sum(axis=0).reindex(df.columns)
        )

    # add sum for numeric cols
    dfout['sum'] = np.nan
    idxs = (dfout['dtype'] == 'float64') | (dfout['dtype'] == 'int64')
    if np.sum(idxs.values) > 0:
        for ft in dfout.loc[idxs].index.values:
            dfout.loc[ft, 'sum'] = df[ft].sum()

    # add min, max for string cols (note the not very clever overwrite of count)
    idxs = (dfout['dtype'] == 'object') | (dfout['dtype'] == 'string[python]')
    if np.sum(idxs.values) > 0:
        for ft in dfout.loc[idxs].index.values:
            dfout.loc[ft, 'min'] = df[ft].value_counts().index.min()
            dfout.loc[ft, 'max'] = df[ft].value_counts().index.max()

    fts_out_all = [
        'dtype',
        'count_null',
        'count_inf',
        'count_zero',
        'count_unique',
        'top',
        'freq',
        'sum',
        'mean',
        'std',
        'min',
        '25%',
        '50%',
        '75%',
        'max',
    ]
    fts_out = [f for f in fts_out_all if f in dfout.columns.values]

    # add mode and mode count WARNING takes forever for large arrays (>10k row)
    if get_mode:
        dfnn = df.select_dtypes(exclude=np.number).copy()
        r = stats.mode(dfnn, axis=0, nan_policy='omit')
        dfmode = pd.DataFrame(
            {'mode': r[0][0], 'mode_count': r[1][0]}, index=dfnn.columns
        )
        dfout = dfout.join(dfmode, how='left', left_index=True, right_index=True)
        fts_out.append(['mode', 'mode_count'])

    # select summary states and prepend random rows for example cases
    rndidx = RNG.integers(low=0, high=len(df), size=nobs)
    dfout = pd.concat(
        (df.iloc[rndidx].T, dfout[fts_out].copy()), axis=1, join='outer', sort=False
    )
    dfout.index.name = 'ft'

    if return_df:
        return dfout
    else:
        display_fw(dfout.iloc[: nfeats + len_idx, :], max_rows=nfeats, **kwargs)
        return f'Shape: {df.shape}, Memsize {nbytes / 1e6:,.1f} MB'


def display_fw(df: pd.DataFrame, **kwargs) -> None:
    """Conv fn: contextually display max rows"""

    options = {
        'display.precision': kwargs.pop('precision', 2),
        'display.max_colwidth': kwargs.pop('max_colwidth', 30),
        'display.max_rows': kwargs.pop('max_rows', 50),
        'display.max_columns': None,
    }

    if kwargs.pop('latex', False):
        options['styler.render.repr'] = 'latex'
        options['styler.latex.environment'] = 'longtable'

    with pd.option_context(*[i for tup in options.items() for i in tup]):
        display(df)


def display_ht(df: pd.DataFrame, **kwargs) -> str:
    """Convenience fn: Display head and tail n rows via display_fw"""

    nrows = kwargs.pop('nrows', 3) if len(df) >= 3 else len(df)
    dfd = df.iloc[np.r_[0:nrows, -nrows:0]].copy()
    display_fw(dfd, **kwargs)
    return f'Shape: {df.shape}, Memsize {df.values.nbytes / 1e6:,.1f} MB'


def get_fts_by_dtype(df: pd.DataFrame, as_dataframe: bool = False) -> dict:
    """Return a dictionary of lists of feats within df according to dtype"""
    fts = dict(
        categorical=[
            k for k, v in df.dtypes.to_dict().items() if v.name[:3] == 'cat'
        ],  # category
        cat=[
            k
            for k, v in df.dtypes.to_dict().items()
            if (v.name[:3] == 'obj') | (v.name[:3] == 'str')
        ],
        bool=[k for k, v in df.dtypes.to_dict().items() if v.name[:3] == 'boo'],
        datetime=[k for k, v in df.dtypes.to_dict().items() if v.name[:3] == 'dat'],
        int=[k for k, v in df.dtypes.to_dict().items() if v.name[:3] == 'int'],
        float=[k for k, v in df.dtypes.to_dict().items() if v.name[:3] == 'flo'],
    )
    w = []
    for _, v in fts.items():
        w += v

    mismatch = list(set(df.columns) - set(w))
    if len(mismatch) > 0:
        raise ValueError(
            f'Failed to match a dtype to {mismatch}'
            + f'\nThese fts did match correctly: {fts}'
        )

    if as_dataframe:
        dtypes = ['categorical', 'cat', 'bool', 'datetime', 'int', 'float']
        d = {w: k for k, v in fts.items() for w in v}
        dfd = pd.DataFrame.from_dict(d, orient='index', columns=['dtype'])
        dfd.index.set_names('ft', inplace=True)
        dfd['dtype'] = pd.Categorical(dfd['dtype'], categories=dtypes, ordered=True)
        return dfd

    return fts



# Copyright 2023 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# curate.data_io.py
"""Data File Handling"""
import csv
import json
import logging
import subprocess
from pathlib import Path

import pandas as pd

from ..utils.file_io import BaseFileIO

__all__ = ['PandasParquetIO', 'PandasCSVIO', 'SimpleStringIO', 'copy_csv2md']

_log = logging.getLogger(__name__)


class PandasParquetIO(BaseFileIO):
    """Helper class to convert pandas to parquet and save to fqn and vice-versa.
    Not strictly needed, but adds a layer of path checking
    """

    def __init__(self, *args, **kwargs):
        """Inherit super"""
        super().__init__(*args, **kwargs)

    def read(self, fn: str, *args, **kwargs) -> pd.DataFrame:
        """Read parquet fn from rootdir, pass args kwargs to pd.read_parquet"""
        fn = Path(fn).with_suffix('.parquet')
        fqn = self.get_path_read(fn)
        _log.info(f'Read from {str(fqn.resolve())}')
        return pd.read_parquet(str(fqn), *args, **kwargs)

    def write(self, df: pd.DataFrame, fn: str, *args, **kwargs) -> Path:
        """Accept pandas DataFrame and fn e.g. `df.parquet`, write to fqn"""
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix('.parquet'))
        df.to_parquet(str(fqn), *args, **kwargs)
        _log.info(f'Written to {str(fqn.resolve())}')
        return fqn


class PandasCSVIO(BaseFileIO):
    """Very simple helper class to write a Pandas dataframe to CSV file in a consistent way"""

    def __init__(self, *args, **kwargs):
        """Inherit super"""
        super().__init__(*args, **kwargs)

    def read(self, fn: str, *args, **kwargs) -> pd.DataFrame:
        """Read csv fn from rootdir, pass args kwargs to pd.read_csv"""
        fn = Path(fn).with_suffix('.csv')
        fqn = self.get_path_read(fn)
        _log.info(f'Read from {str(fqn.resolve())}')
        return pd.read_csv(str(fqn), *args, **kwargs)

    def write(self, df: pd.DataFrame, fn: str, *args, **kwargs) -> str:
        """Accept pandas DataFrame and fn e.g. `df`, write to fn.csv
        Consider using kwarg: float_format='%.3f'
        """
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix('.csv'))
        kws = kwargs.copy()
        kws.update(quoting=csv.QUOTE_NONNUMERIC)
        if (len(df.index.names) == 1) & (df.index.names[0] is None):
            kws.update(index_label='rowid')
        df.to_csv(str(fqn), *args, **kws)
        _log.info(f'Written to {str(fqn.resolve())}')
        return fqn


class SimpleStringIO(BaseFileIO):
    """Helper class to read/write stringlike objects to txt or json files
    Set kind to
        + 'txt' to read/write strings <-> text file
        + 'json' to read/write dicts <-> json file
    """

    def __init__(self, kind: str = 'txt', *args, **kwargs):
        """Init for txt and json only"""
        super().__init__(*args, **kwargs)
        assert kind in set(['txt', 'json']), "kind must be in {'txt', 'json'}"
        self.kind = kind

    def read(self, fn: str) -> str:
        """Read a file from fn according to kind of this object"""
        fqn = self.get_path_read(fn)
        with open(str(fqn), 'r') as f:
            s = f.read().rstrip('\n')
            f.close()
        _log.info(f'Read text from {str(fqn.resolve())}')
        if self.kind == 'json':
            s = json.loads(s)
        return s

    def write(self, s: str, fn: str) -> str:
        fqn = self.get_path_write(Path(self.snl.clean(fn)).with_suffix(f'.{self.kind}'))
        if self.kind == 'json':
            s = json.dumps(s)
        with open(str(fqn), 'w') as f:
            f.write(f'{s}\n')
            f.close()
        _log.info(f'Written to {str(fqn.resolve())}')
        return fqn


def copy_csv2md(fn: str) -> str:
    """Convenience to copy csv 'path/x.csv' to markdown 'path/x.md'"""
    fileio = BaseFileIO()
    fqn = fileio.get_path_read(fn)
    r = subprocess.run(['csv2md', f'{fqn}'], capture_output=True)
    fn_out = f'{fn[:-3] + "md"}'
    fqn_out = fileio.get_path_write(fn_out)
    with open(fqn_out, 'wb') as f:
        f.write(r.stdout)
        f.close()
    _log.info(f'Written to {str(fqn_out.resolve())}')
    return fqn_out



# Copyright 2023 Oreum Industries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model.base.py
"""Base Model Class"""
import logging

import arviz as az
import pymc as pm
import xarray as xa

from .calc import compute_log_likelihood_for_potential

__all__ = ['BasePYMCModel']

_log = logging.getLogger(__name__)
_log_pymc = logging.getLogger('pymc')  # force pymc chatty prints to log
_log_pymc.setLevel(logging.ERROR)
# logging.captureWarnings(True) # further force chatty pymc warnings to log (py.warnings)


class BasePYMCModel:
    """Base handler / wrapper to build, sample, store traces for a pymc model.
    NOTE:
    + This class is to be inherited e.g. `super().__init__(*args, **kwargs)`
    + Children must declare a _build() method and define obs within __init__()
    + To get prior, posterior, and ppc traces, please use the .idata attribute
    """

    rsd = 42

    def __init__(self, **kwargs):
        """Expect obs as dfx pd.DataFrame(mx_en, mx_exs)
        Options for init are often very important!
        https://github.com/pymc-devs/pymc/blob/ed74406735b2faf721e7ebfa156cc6828a5ae16e/pymc/sampling.py#L277
        Usage note: override kws directly on downstream instance e.g.
        def __init__(self, **kwargs):
            super().__init__(*args, **kwargs)
            self.sample_kws.update(dict(tune=1000, draws=500, target_accept=0.85))
        """
        self.model = None
        self._idata = None
        self.replace_idata = False
        self.sample_prior_pred_kws = dict(samples=500, return_inferencedata=True)
        self.sample_post_pred_kws = dict(store_ppc=True, ppc_insample=False)
        self.sample_kws = dict(
            init='auto',  # aka jitter+adapt_diag
            tune=2000,  # twice the 1000 of pymc default
            draws=500,
            chains=4,
            cores=4,
            target_accept=0.8,
            idata_kwargs={"log_likelihood": True},  # safe to have as True
            progressbar=True,
        )
        self.rvs_for_posterior_plots = []
        self.calc_potential_loglike = False
        self.rvs_potential_loglike = None
        self.name = getattr(self, 'name', 'unnamed_model')
        self.version = getattr(self, 'version', 'unversioned_model')
        self.name = f"{self.name}{kwargs.pop('name_ext', '')}"

    @property
    def posterior(self) -> xa.Dataset:
        """Returns posterior from idata from previous run of sample"""
        try:
            self.idata.posterior
        except AttributeError as e:
            raise e("Run sample() first")
        return self.idata.posterior

    @property
    def idata(self) -> az.InferenceData:
        """Returns Arviz InferenceData built from sampling to date"""
        assert self._idata, "Run update_idata() first"
        return self._idata

    def describe_rvs(self) -> dict[list]:
        """Returns a dict of lists of stringnames of RVs"""
        return dict(
            basic=self.model.basic_RVs,
            unobserved=self.model.unobserved_RVs,
            observed=self.model.observed_RVs,
            free=self.model.free_RVs,
            potentials=self.model.potentials,
            deterministics=self.model.deterministics,
        )

    def build(self, **kwargs):
        """Build the model"""
        helper_txt = 'B' if self.model is None else 'Reb'
        try:
            self._build(**kwargs)
            _log.info(f'{helper_txt}uilt model {self.name} {self.version}')
        except AttributeError:
            raise NotImplementedError(
                'Create a method _build() in your'
                + ' subclass, containing the model definition'
            )

    def extend_build(self, **kwargs):
        """Extend build, initially developed to help PPC of GRW and missing value imputation"""
        try:
            self._extend_build(**kwargs)
            _log.info(f'Extended build of model {self.name} {self.version}')
        except AttributeError:
            raise NotImplementedError(
                'Create a method _extend_build() in your'
                + ' subclass, containing the model extension definition'
            )

    def sample_prior_predictive(self, **kwargs):
        """Sample prior predictive:
            use base class defaults self.sample_prior_pred_kws
            or passed kwargs for pm.sample_prior_predictive()
        NOTE:
        + It's not currently possible to run Prior Predictive Checks on a
          model with missing values, per my detailed
          [MRE Notebook gist](https://gist.github.com/jonsedar/070319334bcf033773cc3e9495c79ea0)
          that illustrates the issue.
        + See https://github.com/pymc-devs/pymc/issues/4598
        """
        kws = dict(
            random_seed=kwargs.pop('random_seed', self.rsd),
            samples=kwargs.pop('samples', self.sample_prior_pred_kws['samples']),
            return_inferencedata=kwargs.pop(
                'return_inferencedata',
                self.sample_prior_pred_kws['return_inferencedata'],
            ),
        )
        replace = kwargs.pop('replace', self.replace_idata)

        with self.model:
            try:
                prior_pred = pm.sample_prior_predictive(**{**kws, **kwargs})
            except UserWarning as e:
                _log.warning('Warning in mdl.sample_prior_predictive()', exc_info=e)
            finally:
                _ = self.update_idata(prior_pred, replace=replace)
            _log.info(f'Sampled prior predictive for {self.name} {self.version}')
        return None

    def sample(self, **kwargs):
        """Sample posterior: use base class defaults self.sample_kws
        or passed kwargs for pm.sample()
        """
        kws = dict(
            init=kwargs.pop('init', self.sample_kws['init']),
            random_seed=kwargs.pop('random_seed', self.rsd),
            tune=kwargs.pop('tune', self.sample_kws['tune']),
            draws=kwargs.pop('draws', self.sample_kws['draws']),
            chains=kwargs.pop('chains', self.sample_kws['chains']),
            cores=kwargs.pop('cores', self.sample_kws['cores']),
            progressbar=kwargs.pop('progressbar', self.sample_kws['progressbar']),
            idata_kwargs=kwargs.pop('idata_kwargs', self.sample_kws['idata_kwargs']),
        )

        target_accept = kwargs.pop('target_accept', self.sample_kws['target_accept'])
        step = kwargs.pop('step', 'nuts')

        with self.model:
            common_stepper_options = {
                'nuts': pm.NUTS(target_accept=target_accept),
                'metropolis': pm.Metropolis(target_accept=target_accept),
                'advi': pm.ADVI(),
            }
            kws['step'] = common_stepper_options.get(step, None)

            try:
                posterior = pm.sample(**{**kws, **kwargs})
            except UserWarning as e:
                _log.warning('Warning in mdl.sample()', exc_info=e)
            finally:
                _ = self.update_idata(posterior)

        _log.info(f'Sampled posterior for {self.name} {self.version}')

        # optional calculate loglikelihood for potentials
        if self.calc_potential_loglike:
            self.idata.add_groups(
                dict(
                    log_likelihood=compute_log_likelihood_for_potential(
                        idata=self.idata,
                        model=self.model,
                        var_names=self.rvs_potential_loglike,
                        extend_inferencedata=False,
                    )
                )
            )
            # rename to have exact same name as observedRVs
            for nm in self.rvs_potential_loglike:
                nm0 = nm.lstrip('pot_')
                self.idata['log_likelihood'][nm0] = self.idata['log_likelihood'][nm]
                del self.idata['log_likelihood'][nm]

        return None

    def sample_posterior_predictive(self, **kwargs):
        """Sample posterior predictive
        use self.sample_post_pred_kws or passed kwargs
        Note by default aimed toward out-of-sample PPC in production
        """
        store_ppc = kwargs.pop('store_ppc', self.sample_post_pred_kws['store_ppc'])
        kws = dict(
            trace=self.posterior,
            random_seed=kwargs.pop('random_seed', self.rsd),
            predictions=not kwargs.pop(
                'ppc_insample', self.sample_post_pred_kws['ppc_insample']
            ),
        )
        with self.model:
            ppc = pm.sample_posterior_predictive(**{**kws, **kwargs})

        _log.info(f'Sampled ppc for {self.name} {self.version}')

        if store_ppc:
            _ = self.update_idata(ppc)
        else:
            return ppc

    def replace_obs(self, obsd: dict = None) -> None:
        """Replace the observations
        Assumes data lives in pm.MutableData containers in your _build() function
        You must call `build()` afterward
        Optionally afterwards call `extend_build()` for future time-dependent PPC
        """
        for k, v in obsd.items():
            setattr(self, k, v)
            _log.info(f'Replaced obs {k} in {self.name} {self.version}')

    def update_idata(self, idata: az.InferenceData, replace: bool = False) -> None:
        """Create (and update) an Arviz InferenceData object on-model from a
        passed-in presampled InferenceData object
        """
        # TODO improve this logic to use self.idata
        if self._idata is None:
            self._idata = idata
        else:
            side = 'right' if replace else 'left'
            self._idata.extend(idata, join=side)
